"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createTaskRunner = void 0;
const cron_parser_1 = require("cron-parser");
const path_1 = require("path");
const tiny_glob_1 = __importDefault(require("tiny-glob"));
async function loadTasks(folder, pattern) {
    if (!(0, path_1.isAbsolute)(folder))
        throw new Error('folder must be an absolute path');
    const files = await (0, tiny_glob_1.default)(pattern, { cwd: folder });
    const tasks = [];
    await Promise.all(files.map(async (file) => {
        const module = require((0, path_1.join)(folder, file).replace(/\\/g, '/'));
        const isValid = 'default' in module && typeof module.default === 'object';
        if (!isValid) {
            console.error('Error loading task file:', file, 'it must contain single default export:');
            console.error('     ```````````````````````````````````````````````````````````');
            console.error("     ` const task = taskBuilder.daily().at('11:30').build(); `");
            console.error('     ` export default task                                     `');
            console.error('     ```````````````````````````````````````````````````````````');
            return;
        }
        tasks.push({ ...module.default, path: file });
    }));
    console.info('Loaded', tasks.length, 'tasks from folder', folder);
    // TODO: use tiny-glob
    return tasks;
}
function shouldRunTask(task, tickTime) {
    if (task.runsAt) {
        // [tick1] ... [tick2] <->. [runsAt] ....
        if (task._executed)
            return;
        if (Date.now() > task.runsAt.getTime() - tickTime) {
            return true;
        }
    }
    else {
        const cron = (0, cron_parser_1.parseExpression)(task.cron);
        const nextSuitable = cron.next();
        if (Date.now() > nextSuitable.toDate().getTime() - tickTime) {
            return true;
        }
    }
    return false;
}
function taskToString(task) {
    if (task.name)
        return `${task.name} (${task.cron || task.runsAt?.toLocaleString()})`;
    if (task.path)
        return `${task.path} (${task.cron || task.runsAt?.toLocaleString()})`;
    return `Unknown task (${task.cron || task.runsAt?.toLocaleString()})`;
}
async function createTaskRunner(options) {
    const tasks = options.tasks || (await loadTasks(options.folder, options.pattern));
    const tickTime = options.tickTime || 1000;
    const tick = () => {
        tasks.forEach(task => {
            if (!shouldRunTask(task, tickTime))
                return;
            if (task.runsAt) {
                task._executed = true;
            }
            try {
                task.execute()
                    .then(() => {
                    options.onTaskComplete?.(task);
                })
                    .catch(err => {
                    console.error('\x1b[31m[Error] at task', taskToString(task), '\x1b[0m');
                    console.error(err);
                });
            }
            catch (err) {
                console.error('Error running task', taskToString(task), err);
            }
        });
    };
    const tid = setInterval(tick, tickTime);
    const stop = () => clearInterval(tid);
    return { tid, stop };
}
exports.createTaskRunner = createTaskRunner;
